from typing import Union


class Host:
    hostname: str
    port: Union[int, None]
    username: Union[str, None]
    password: Union[str, None]

    def __init__(
        self, 
        netloc: str
    ):
        userinfo, _, host = netloc.rpartition('@')
        if host.count(':') == 1 or '[' in host:
            hostname, _, port = host.rpartition(':')
            port = int(port)
        else:
            hostname, port = host, None
        if hostname.startswith('[') and hostname.endswith(']'):
            hostname = hostname[1:-1]
        if userinfo:
            username, _, password = userinfo.partition(':')
        else:
            username = password = None

        self.netloc = netloc
        self.hostname = hostname
        self.port = port
        self.username = username
        self.password = password

    @property
    def host(self):
        host = f'[{self.hostname}]' if ':' in self.hostname else self.hostname
        if self.port:
            host = f'{host}:{self.port}'
        return host

    def __str__(self):
        userinfo = ''
        if self.username:
            userinfo += self.username
            if self.password:
                userinfo += ':' + self.password
            userinfo += '@'
        return userinfo + self.host



from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    IPvAnyAddress
)

from typing import (
    Dict, 
    Tuple,
    Literal
)


class Service(BaseModel):
    service_instance: StrictStr
    service_name: StrictStr
    service_protocol: Literal["udp", "tcp"]
    service_url: StrictStr
    service_ip: IPvAnyAddress
    service_port: StrictInt
    service_context: Dict[StrictStr, StrictStr]={}

    def to_address(self) -> Tuple[str, int]:
        return (
            str(self.service_ip),
            self.service_port
        )


import io
import struct
import time
from .query_type import QueryType
from typing import Dict, Tuple
from .record_data_types import (
    ARecordData,
    AAAARecordData,
    CNAMERecordData,
    MXRecordData,
    NAPTRRecordData,
    NSRecordData,
    PTRRecordData,
    RecordData,
    RecordType,
    RecordTypesMap,
    SOARecordData,
    SRVRecordData,
    TXTRecordData,
    UnsupportedRecordData

)
from mercury_sync.discovery.dns.core.record.record_data_types.utils import (
    load_domain_name, 
    pack_domain_name, 
    pack_string
)


MAXAGE = 3600000



class Record:

    record_types: Dict[
        RecordType,
        RecordData
    ] = {
        RecordType.A: ARecordData,
        RecordType.AAAA: AAAARecordData,
        RecordType.CNAME: CNAMERecordData,
        RecordType.MX: MXRecordData,
        RecordType.NAPTR: NAPTRRecordData,
        RecordType.NS: NSRecordData,
        RecordType.PTR: PTRRecordData,
        RecordType.SOA: SOARecordData,
        RecordType.SRV: SRVRecordData,
        RecordType.TXT: TXTRecordData
    }

    def __init__(
        self,
        query_type: QueryType = QueryType.REQUEST,
        name: str = '',
        record_type: RecordType = RecordType.ANY,
        qclass: int = 1,
        ttl: int = 0,
        data: Tuple[int, RecordData] = None
    ):
        self.query_type = query_type
        self.name = name
        self.record_type = record_type
        self.qclass = qclass

        self.ttl = ttl  # 0 means item should not be cached
        self.data = data
        self.timestamp = int(time.time())

        self.types_map = RecordTypesMap()

    @classmethod
    def create_rdata(
        cls,
        record_type: RecordType, 
        *args
    ) -> RecordData:
        
        record_data = cls.record_types.get(record_type)

        if record_data is None:
            return UnsupportedRecordData(
                record_type,
                *args
            )

        return record_data(
            *args
        )

    @classmethod
    def load_rdata(
        cls,
        record_type: RecordType, 
        data: bytes, 
        cursor_position: int,
        size: int
    ) -> Tuple[int, RecordData]:
        '''Load RData from a byte sequence.'''
        record_data = cls.record_types.get(record_type)
        if record_data is None:
            return UnsupportedRecordData.load(
                data, 
                cursor_position, 
                size, 
                record_type
            )
        
        return record_data.load(
            data, 
            cursor_position, 
            size
        )

    def copy(
        self, 
        **kwargs
    ):
        return Record(
            query_type=kwargs.get('query_type', self.query_type),
            name=kwargs.get('name', self.name),
            record_type=kwargs.get('record_type', self.record_type),
            qclass=kwargs.get('qclass', self.qclass),
            ttl=kwargs.get('ttl', self.ttl),
            data=kwargs.get('data', self.data)
        )

    def parse(
        self, 
        data: bytes, 
        cursor_position: int
    ):
        cursor_position, self.name = load_domain_name(
            data, 
            cursor_position
        )

        record_type, self.qclass = struct.unpack(
            '!HH', 
            data[cursor_position:cursor_position + 4]
        )

        self.record_type = self.types_map.types_by_code.get(
            record_type
        )

        cursor_position += 4
        if self.query_type == QueryType.RESPONSE:
            self.timestamp = int(time.time())
            self.ttl, size = struct.unpack(
                '!LH', 
                data[cursor_position:cursor_position + 6]
            )

            cursor_position += 6

            _, self.data = Record.load_rdata(
                self.record_type, 
                data, 
                cursor_position, 
                size
            )

            cursor_position += size

        return cursor_position

    def pack(
        self, 
        names, 
        offset=0
    ):
        buf = io.BytesIO()

        buf.write(
            pack_domain_name(
                self.name, 
                names, 
                offset
            )
        )

        buf.write(
            struct.pack(
                '!HH', 
                self.record_type.value, 
                self.qclass
            )
        )

        if self.query_type == QueryType.RESPONSE:

            if self.ttl < 0:
                ttl = MAXAGE

            else:
                now = int(time.time())
                self.ttl -= now - self.timestamp

                if self.ttl < 0:
                    self.ttl = 0

                self.timestamp = now
                ttl = self.ttl

            buf.write(
                struct.pack(
                    '!L', 
                    ttl
                )
            )

            data_str = b''.join(
                self.data.dump(
                    names, 
                    offset + buf.tell()
                )
            )

            buf.write(
                pack_string(
                    data_str, 
                    '!H'
                )
            )

        return buf.getvalue()




import io
import base64
import struct
from mercury_sync.discovery.dns.core.exceptions import (
    DNSError
)
from mercury_sync.discovery.dns.core.record import (
    QueryType,
    Record,
    RecordType
)
from pydantic import (
    StrictInt, 
    StrictBool
)
from typing import (
    List, 
    Tuple,
    Union,
    Iterable,
    Dict,
    Optional
)
from .http_request import HTTPRequest, HTTPRequestMethod
from .message import Message




class DNSMessage(Message):
    query_type: QueryType=QueryType.REQUEST
    query_id: StrictInt=0
    query_opcode: StrictInt=0
    query_authoritative_answer: StrictInt=0
    query_truncation: StrictInt=0
    query_desired_recursion: StrictInt=0
    query_available_recursion: StrictInt=0
    query_result_code: StrictInt=0
    record_types: List[RecordType]=[]
    query_domains: List[Record]=[]
    query_answers: List[Record]=[]
    query_namservers: List[Record]=[]
    query_additional_records: List[Record]=[]
    query_has_result: StrictBool=False

    class Config:
        arbitrary_types_allowed=True


    def __iter__(self):
        return iter(self.query_answers)
    
    def is_request(self):
        return self.query_type
    
    @classmethod
    def get_bits(
        cls,
        num: int, 
        bit_len: int
    ):

        high = num >> bit_len
        low = num - (high << bit_len)

        return low, high
    
    @staticmethod
    def parse_entry(
        query_type: QueryType, 
        data: bytes, 
        cursor_posiition: int,
        length: int
    ) -> Tuple[int, List[Record]]:
        
        results: List[Record] = []

        for _ in range(length):
            record = Record(query_type.value)
            cursor_posiition = record.parse(data, cursor_posiition)

            results.append(record)

        return cursor_posiition, results

    @classmethod
    def parse(
        cls, 
        data: bytes, 
        query_id: Optional[bytes] = None
    ):

        (
            request_id, 
            raw_data, 
            domains, 
            answers, 
            nameservers, 
            additional_records
        ) = struct.unpack('!HHHHHH', data[:12])

        if query_id is not None and query_id != request_id:
            raise DNSError(-1, 'Transaction ID mismatch')
        
        result_code, raw_data = cls.get_bits(raw_data, 4)  # rcode: 0 for no error

        _, raw_data = cls.get_bits(raw_data, 3)  # reserved

        available_recursion, raw_data = cls.get_bits(raw_data, 1)  # recursion available

        desired_recursion, raw_data = cls.get_bits(raw_data, 1)  # recursion desired

        truncation, raw_data = cls.get_bits(raw_data, 1)  # truncation

        authoritative_answer, raw_data = cls.get_bits(raw_data, 1)  # authoritative answer

        opcode, raw_data = cls.get_bits(raw_data, 4)  # opcode
        
        query_type, raw_data = cls.get_bits(raw_data, 1)  # qr: 0 for query and 1 for response

        cursor_position, query_domains = cls.parse_entry(
            QueryType.REQUEST.value, 
            data, 
            12, 
            domains
        )

        cursor_position, query_answers = cls.parse_entry(
            QueryType.RESPONSE.value, 
            data, 
            cursor_position, 
            answers
        )

        cursor_position, query_nameservers = cls.parse_entry(
            QueryType.RESPONSE.value, 
            data, 
            cursor_position, 
            nameservers
        )

        _, query_additional_records = cls.parse_entry(
            QueryType.RESPONSE.value, 
            data, 
            cursor_position, 
            additional_records
        )

        return DNSMessage(
            query_type=QueryType.by_value(query_type),
            query_opcode=opcode,
            query_authoritative_answer=authoritative_answer,
            query_truncation=truncation,
            query_desired_recursion=desired_recursion,
            query_available_recursion=available_recursion,
            query_result_code=result_code,
            query_domains=query_domains,
            query_answers=query_answers,
            query_namservers=query_nameservers,
            query_additional_records=query_additional_records
        )

    def get_record(
        self, 
        record_types: Union[RecordType, Iterable[RecordType]]
    ):
        '''Get the first record of qtype defined in `qtypes` in answer list.
        '''
        if isinstance(record_types, RecordType):
            record_types = record_types

        for item in self.query_answers:
            if item.record_types in record_types:
                return item.data
            
    def pack(
        self, 
        size_limit: int = None
    ) -> bytes:


        names: Dict[str, int] = {}
        buffer = io.BytesIO()
        buffer.seek(12)
        truncation = 0

        query_groups = [
            self.query_domains, 
            self.query_answers, 
            self.query_namservers, 
            self.query_additional_records
        ]

        for group in query_groups:
            if truncation: 
                break

            for record in group:
                offset = buffer.tell()
                packed_record = record.pack(names, offset)

                if size_limit is not None and offset + len(packed_record) > size_limit:
                    truncation = 1
                    break

                buffer.write(packed_record)

        self.query_truncation = truncation
        buffer.seek(0)

        query_type = self.query_type.value << 15
        query_opcode = self.query_opcode << 11
        query_authoritative_answer = self.query_authoritative_answer << 10
        query_truncation = truncation << 9
        query_desired_recursion = self.query_desired_recursion << 8
        query_available_recursion = self.query_available_recursion << 7
        query_buffer_extra = 0 << 4
        query_result_code = self.query_result_code

        query_data = sum([
            query_type,
            query_opcode,
            query_authoritative_answer,
            query_truncation,
            query_desired_recursion,
            query_available_recursion,
            query_buffer_extra,
            query_result_code
        ])



        buffer.write(
            struct.pack(
                '!HHHHHH', 
                self.query_id, 
                query_data,
                len(self.query_domains),
                len(self.query_answers), 
                len(self.query_namservers), 
                len(self.query_additional_records)
            )
        )

        return buffer.getvalue()
    
    def to_http_bytes(
        self,
        url: str,
        method: HTTPRequestMethod=HTTPRequestMethod.GET
    ) -> bytes:
        
        message = self.pack()
        params: Dict[str, str] = {}
        data: Union[str, None] = None

        if method == HTTPRequestMethod.GET:
            params['dns'] = base64.urlsafe_b64encode(message).decode().rstrip('=')

        else:
            data = message.decode()
            
        http_request = HTTPRequest(
            host=self.host,
            port=self.port,
            error=self.error,
            url=url,
            method=method,
            headers={
                'accept': 'application/dns-message',
                'content-type': 'application/dns-message',
            },
            data=data
        )

        return http_request.prepare_request()
    
    def to_tcp_bytes(self) -> Tuple[bytes, bytes]:
        message = self.pack()
        message_size = len(message)

        return struct.pack('!H', message_size), + message
    
    def to_udp_bytes(self) -> bytes:
        return self.pack()



from mercury_sync.discovery.dns.core.record.record_data_types import (
    RecordData
)
from mercury_sync.discovery.dns.core.record import (
    Record, 
    RecordType
)
from typing import Dict, Iterable, Union
from .cache_value import CacheValue


class CacheNode:
    def __init__(self):
        self.children: Dict[str, CacheNode] = {}
        self.data = CacheValue()

    def get(self, fqdn: str, touch: bool = False):
        current = self
        keys = reversed(fqdn.split('.'))
        for key in keys:
            child = current.children.get(key)

            if child is None:
                child = current.children.get('*')

            if child is None and touch is False:
                return None

            elif child is None and touch:
                child = CacheNode()
                current.children[key] = child
                
            current = child
        return current.data

    def query(
        self, 
        fqdn: str, 
        record_type: Union[RecordType, Iterable[RecordType]]
    ):
        if isinstance(record_type, RecordType):
            value = self.get(fqdn)
            if value is not None:
                yield from value.get(record_type)
        else:
            for rtype in record_type:
                yield from self.query(fqdn, rtype)

    def add(
        self,
        fqdn: str = None,
        record_type: RecordType = None,
        data: Union[RecordData, bytes, Iterable] = None,
        ttl=-1,
        record: Record = None
    ):
        if record is None:

            if isinstance(data, bytes):
                _, rdata = Record.load_rdata(record_type, data, 0, len(data))

            elif isinstance(data, RecordData):
                rdata = data

            else:
                rdata = Record.create_rdata(record_type, *data)

            record = Record(
                name=fqdn, 
                data=rdata, 
                record_type=record_type, 
                ttl=ttl
            )

        value = self.get(record.name, True)
        value.add(record)

    def iter_values(self) -> Iterable[Record]:

        yield from self.data.get(RecordType.ANY)

        for child in self.children.values():
            yield from child.iter_values()



from __future__ import annotations
from typing import Dict, Iterable, Tuple, Optional
from .record_data import RecordData
from .record_types import RecordType
from .utils import (
    pack_domain_name
)


class DomainRecordData(RecordData):
    '''A record'''

    def __init__(
        self, 
        record_type: RecordType,
        data: Optional[str]=None
    ):
        super().__init__(
            record_type,
            data=data
        )

    @classmethod
    def load(
        cls, 
        data: bytes, 
        cursor_position: int,
        size: int
    ) -> Tuple[int, DomainRecordData]:
        raise NotImplementedError('Err. - Not implemented for DomainRecordData type')

    def dump(
        self, 
        names: Dict[str, int], 
        offset: int
    ) -> Iterable[bytes]:
        yield pack_domain_name(self.data, names, offset + 2)





from __future__ import annotations
from typing import Tuple, Iterable, Dict
from .record_data import RecordData
from .record_types import RecordType


class UnsupportedRecordData(RecordData):
    '''Unsupported RData'''
    def __init__(self, rtype: RecordType, raw: str):

        super().__init__(
            rtype,
            data=raw.encode()
        )

        self.raw = raw

    @classmethod
    def load(
        cls, 
        data: bytes, 
        cursor_position: int, 
        size: int,
        record_type: RecordType
    ) -> Tuple[int, UnsupportedRecordData]:
        
        return cursor_position + size, UnsupportedRecordData(
            record_type, 
            data[cursor_position:cursor_position + size]
        )

    def dump(self, names: Dict[str, int], offset: int) -> Iterable[bytes]:
        yield self.raw



from __future__ import annotations
import asyncio
import pickle
import socket
import ssl
import zstandard
from collections import deque, defaultdict
from dtls import do_patch
from mercury_sync.connection.base.connection_type import ConnectionType
from mercury_sync.connection.udp.protocols import MercurySyncUDPProtocol
from mercury_sync.encryption import AESGCMFernet
from mercury_sync.env import Env
from mercury_sync.env.time_parser import TimeParser
from mercury_sync.models.message import Message
from mercury_sync.snowflake.snowflake_generator import SnowflakeGenerator
from typing import (
    Tuple, 
    Deque, 
    Any, 
    Dict, 
    Coroutine, 
    AsyncIterable,
    Optional,
    Union
)

do_patch()


class MercurySyncUDPConnection:

    def __init__(
        self,
        host: str,
        port: int,
        instance_id: int,
        env: Env
    ) -> None:

        self.id_generator = SnowflakeGenerator(instance_id)
        self.env = env

        self.host = host
        self.port = port

        self.events: Dict[
            str,
            Coroutine
        ] = {}

        self._transport: asyncio.DatagramTransport = None
        self._loop: Union[asyncio.AbstractEventLoop, None] = None
        self.queue: Dict[str, Deque[Tuple[str, int, float, Any]] ] = defaultdict(deque)
        self.parsers: Dict[str, Message] = {}
        self._waiters: Dict[str, Deque[asyncio.Future]] = defaultdict(deque)
        self._pending_responses: Deque[asyncio.Task] = deque()

        self._udp_cert_path: Union[str, None] = None
        self._udp_key_path: Union[str, None] = None
        self._udp_ssl_context: Union[ssl.SSLContext, None] = None

        self._encryptor = AESGCMFernet(env)
        self._semaphore: Union[asyncio.Semaphore, None] = None
        self._compressor: Union[zstandard.ZstdCompressor, None] = None
        self._decompressor: Union[zstandard.ZstdDecompressor, None] = None
        
        self._running = False
        self._cleanup_task: Union[asyncio.Task, None] = None
        self._sleep_task: Union[asyncio.Task, None] = None
        self._cleanup_interval = TimeParser(env.MERCURY_SYNC_CLEANUP_INTERVAL).time
        self._max_concurrency = env.MERCURY_SYNC_MAX_CONCURRENCY
        self.udp_socket: Union[socket.socket, None] = None

        self.connection_type = ConnectionType.UDP

    def connect(
        self, 
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None,
        worker_socket: Optional[socket.socket]=None
    ) -> None:
        
        try:

            self._loop = asyncio.get_event_loop()

        except Exception:
            self._loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._loop)

        self._running = True

        self._semaphore = asyncio.Semaphore(self._max_concurrency)

        self._compressor = zstandard.ZstdCompressor()
        self._decompressor = zstandard.ZstdDecompressor()

        if worker_socket is None:
            self.udp_socket = socket.socket(
                socket.AF_INET, 
                socket.SOCK_DGRAM, 
                socket.IPPROTO_UDP
            )

            self.udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.udp_socket.bind((
                self.host,
                self.port
            ))

            self.udp_socket.setblocking(False)

        else:
            self.udp_socket = worker_socket
            host, port = self.udp_socket.getsockname()
            
            self.host = host
            self.port = port

        if cert_path and key_path:
            self._udp_ssl_context = self._create_udp_ssl_context(
                cert_path=cert_path,
                key_path=key_path,
            )

            self.udp_socket = self._udp_ssl_context.wrap_socket(self.udp_socket)

        server = self._loop.create_datagram_endpoint(
            lambda: MercurySyncUDPProtocol(
                self.read
            ),
            sock=self.udp_socket
        )

        transport, _ = self._loop.run_until_complete(server)
        self._transport = transport

        self._cleanup_task = self._loop.create_task(self._cleanup())

    async def connect_async(
        self, 
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None,
        worker_socket: Optional[socket.socket]=None
    ) -> None:
        
        self._loop = asyncio.get_event_loop()
        self._running = True

        self._semaphore = asyncio.Semaphore(self._max_concurrency)

        self._compressor = zstandard.ZstdCompressor()
        self._decompressor = zstandard.ZstdDecompressor()

        if worker_socket is None:
            self.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            self.udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.udp_socket.bind((
                self.host,
                self.port
            ))

            self.udp_socket.setblocking(False)

        else:
            self.udp_socket = worker_socket


        if cert_path and key_path:
            self._udp_ssl_context = self._create_udp_ssl_context(
                cert_path=cert_path,
                key_path=key_path,
            )

            self.udp_socket = self._udp_ssl_context.wrap_socket(self.udp_socket)

        server = self._loop.create_datagram_endpoint(
            lambda: MercurySyncUDPProtocol(
                self.read
            ),
            sock=self.udp_socket
        )

        transport, _ = await server
        self._transport = transport

        self._cleanup_task = self._loop.create_task(self._cleanup())

    def _create_udp_ssl_context(
        self,
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None
    ) -> ssl.SSLContext: 
        
        if self._udp_cert_path is None:
            self._udp_cert_path = cert_path

        if self._udp_key_path is None:
            self._udp_key_path = key_path

        ssl_ctx = ssl.SSLContext(ssl.PROTOCOL_TLS)
        ssl_ctx.options |= ssl.OP_NO_TLSv1
        ssl_ctx.options |= ssl.OP_NO_TLSv1_1
        ssl_ctx.options |= ssl.OP_SINGLE_DH_USE
        ssl_ctx.options |= ssl.OP_SINGLE_ECDH_USE
        ssl_ctx.load_cert_chain(cert_path, keyfile=key_path)
        ssl_ctx.load_verify_locations(cafile=cert_path)
        ssl_ctx.check_hostname = False
        ssl_ctx.verify_mode = ssl.VerifyMode.CERT_REQUIRED
        ssl_ctx.set_ciphers('ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384')
        
        return ssl_ctx
    
    async def _cleanup(self):
        while self._running:
            self._sleep_task = asyncio.create_task(
                asyncio.sleep(self._cleanup_interval)
            )

            await self._sleep_task

            for pending in list(self._pending_responses):
                if pending.done() or pending.cancelled():
                    try:
                        await pending

                    except (
                        ConnectionAbortedError,
                        ConnectionRefusedError,
                        ConnectionResetError,
                        asyncio.CancelledError,
                        asyncio.InvalidStateError
                    ):
                        await self.close()
                        await self.connect_async(
                            cert_path=self._udp_cert_path,
                            key_path=self._udp_key_path
                        )

                    self._pending_responses.pop()
    
    async def send(
        self, 
        event_name: str,
        data: Any, 
        addr: Tuple[str, int]
    ) -> Tuple[int, Dict[str, Any]]:

        item = pickle.dumps((
            'request',
            self.id_generator.generate(),
            event_name,
            data
        ), protocol=pickle.HIGHEST_PROTOCOL)

        encrypted_message = self._encryptor.encrypt(item)
        compressed = self._compressor.compress(encrypted_message)
        
        self._transport.sendto(compressed, addr)

        waiter = self._loop.create_future()
        self._waiters[event_name].append(waiter)

        (
            _,
            shard_id,
            _,
            response_data,
            _, 
            _
        ) = await waiter

        return (
            shard_id,
            response_data
        )
    
    async def send_bytes(
        self,
        event_name: str,
        data: bytes,
        addr: Tuple[str, int]
    ) -> bytes:
        
        self._transport.sendto(data, addr)

        waiter = self._loop.create_future()
        self._waiters[event_name].append(waiter)
        return await waiter

    async def stream(
        self, 
        event_name: str,
        data: Any, 
        addr: Tuple[str, int]
    ) -> AsyncIterable[Tuple[int, Dict[str, Any]]]: 

        item = pickle.dumps((
            'stream',
            self.id_generator.generate(),
            event_name,
            data
        ), protocol=pickle.HIGHEST_PROTOCOL)

        encrypted_message = self._encryptor.encrypt(item)
        compressed = self._compressor.compress(encrypted_message)
        
        self._transport.sendto(compressed, addr)

        waiter = self._loop.create_future()
        self._waiters[event_name].append(waiter)

        await waiter

        for item in self.queue[event_name]:
            (
                _,
                shard_id,
                _,
                response_data,
                _, 
                _
            ) = item

            yield(
                shard_id,
                response_data
            )

        self.queue.clear()

    def read(
        self,
        data: bytes, 
        addr: Tuple[str, int]
    ) -> None:
        
        decrypted = self._encryptor.decrypt(
            self._decompressor.decompress(data)
        )

        result: Tuple[
            str, 
            int, 
            float, 
            Any
        ] = pickle.loads(decrypted)

        (
            message_type, 
            shard_id, 
            event_name, 
            payload
        ) = result

        incoming_host, incoming_port = addr

        if message_type == 'request':
            self._pending_responses.append(
                asyncio.create_task(
                    self._read(
                        event_name,
                        self.events.get(event_name)(
                            shard_id,
                            self.parsers[event_name](**payload)
                        ),
                        addr
                    )
                )
            )
            
        elif message_type == 'stream':
            self._pending_responses.append(
                asyncio.create_task(
                    self._read_iterator(
                        event_name,
                        self.events.get(event_name)(
                            shard_id,
                            self.parsers[event_name](**payload)
                        ),
                        addr
                    )
                )
            )

        else:

            event_waiter = self._waiters[event_name]
      
            if bool(event_waiter):
                waiter = event_waiter.pop()
                
                try:
                    
                    waiter.set_result((
                        message_type, 
                        shard_id,
                        event_name,
                        payload, 
                        incoming_host,
                        incoming_port
                    ))
                
                except asyncio.InvalidStateError:
                    pass


    async def _read(
        self,
        event_name: str,
        coroutine: Coroutine,
        addr: Tuple[str, int]
    ) -> Coroutine[Any, Any, None]:
        response: Message = await coroutine

        item = pickle.dumps(
            (
                'response', 
                self.id_generator.generate(),
                event_name,
                response.to_data()
            ),
            protocol=pickle.HIGHEST_PROTOCOL
        )

        encrypted_message = self._encryptor.encrypt(item)
        compressed = self._compressor.compress(encrypted_message)

        self._transport.sendto(compressed, addr)

    async def _read_iterator(
        self,
        event_name: str,
        coroutine: AsyncIterable[Message],
        addr: Tuple[str, int]    
    ) -> Coroutine[Any, Any, None]:
        async for response in coroutine:

            item = pickle.dumps(
                (
                    'response', 
                    self.id_generator.generate(),
                    event_name,
                    response.to_data()
                ),
                protocol=pickle.HIGHEST_PROTOCOL
            )

            encrypted_message = self._encryptor.encrypt(item)
            compressed = self._compressor.compress(encrypted_message)
            self._transport.sendto(compressed, addr)

    async def close(self) -> None:
        self._running = False
        self._transport.abort()
        
        if self._cleanup_task:
            self._cleanup_task.cancel()
            if self._cleanup_task.cancelled() is False:
                try:
                    self._sleep_task.cancel()
                    if not self._sleep_task.cancelled():
                        await self._sleep_task

                except asyncio.CancelledError:
                    pass

                try:

                    await self._cleanup_task

                except asyncio.CancelledError:
                    pass


import asyncio
from mercury_sync.discovery.dns.core.record import (
    RecordType,
    RecordTypesMap
)
from mercury_sync.discovery.dns.core.record.record_data_types import RecordData
from mercury_sync.discovery.dns.core.url import URL
from mercury_sync.env import Env
from mercury_sync.models.dns_message import DNSMessage
from typing import Literal, Optional, List, Tuple, Callable, Union
from .proxy_resolver import ProxyResolver
from .recursive_resolver import RecursiveResolver


Proxy = List[
    Tuple[
        Union[
            Callable[
                [str],
                bool
            ],
            str,
            None
        ], 
        str
    ]
]


class DNSResolver:

    def __init__(
        self,
        host: str,
        port: int,
        instance_id: str,
        env: Env,
        resolver: Literal["proxy", "recursive"]="proxy",
        proxies: Optional[
            List[Proxy]
        ]=None
    ) -> None:
        
        if resolver == "proxy":
            self.resolver = ProxyResolver(
                host,
                port,
                instance_id,
                env,
                proxies=proxies
            )

        else:
            self.resolver = RecursiveResolver(
                host,
                port,
                instance_id,
                env
            )

        self.types_map = RecordTypesMap()

    def add_to_cache(
        self,
        domain: str,
        record_type: RecordType,
        data: RecordData,
        ttl: Union[int, float]=-1
    ):
        self.resolver.cache.add(
            fqdn=domain,
            record_type=record_type,
            data=data,
            ttl=ttl
        )

    def add_nameservers(
        self,
        urls: List[str]
    ):
        return self.resolver.add_nameserver(urls)

    def set_proxies(
        self,
        proxies: List[Proxy]
    ):
        if isinstance(self.resolver, ProxyResolver):
            self.resolver.set_proxies(proxies)

    def download_common(self):
        if isinstance(self.resolver, RecursiveResolver):
            self.resolver.load_nameserver_cache()

    async def connect_nameservers(
        self,
        urls: List[URL],
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None,
    ):
        
        await asyncio.gather(*[
            self.resolver.client.connect_client(
                url,
                cert_path=cert_path,
                key_path=key_path
            ) for url in urls
        ])

    async def query(
        self,
        domain_name: str,
        record_type: RecordType=RecordType.SRV,
        skip_cache: bool=False
    ) -> Tuple[DNSMessage, bool]:

        try:
            result = await self.resolver.query(
                domain_name,
                record_type=record_type,
                skip_cache=skip_cache
            )

            return result, True
        
        except asyncio.TimeoutError:
            return DNSMessage(), False


'''
This module load nameservers from Windows Registry.
'''

import winreg


def _nt_read_key(hlm, key):
    
    regkey = winreg.OpenKey(hlm, key)

    try:
        value, _rtype = winreg.QueryValueEx(regkey, 'NameServer')
        if not value:
            value, _rtype = winreg.QueryValueEx(regkey, 'DhcpNameServer')
    except Exception:
        value = None
    regkey.Close()
    if value:
        sep = ',' if ',' in value else ' '
        return value.split(sep)


def _nt_is_enabled(hlm, guid):
    connection_key = winreg.OpenKey(
        hlm,
        r'SYSTEM\CurrentControlSet\Control\Network\{4D36E972-E325-11CE-BFC1-08002BE10318}\%s\Connection'
        % guid)
    pnp_id, _ttype = winreg.QueryValueEx(connection_key, 'PnpInstanceID')
    device_key = winreg.OpenKey(hlm,
                                r'SYSTEM\CurrentControlSet\Enum\%s' % pnp_id)
    try:
        flags, _ttype = winreg.QueryValueEx(device_key, 'ConfigFlags')
        return not flags & 0x1
    except Exception:
        return False
    finally:
        device_key.Close()
        connection_key.Close()


def get_nameservers():
    '''
    Get nameservers from Windows Registry.
    '''
    nameservers = []
    hlm = winreg.ConnectRegistry(None, winreg.HKEY_LOCAL_MACHINE)
    servers = _nt_read_key(
        hlm, r'SYSTEM\CurrentControlSet\Services\Tcpip\Parameters')
    if servers is not None:
        nameservers.extend(servers)
    interfaces = winreg.OpenKey(
        hlm, r'SYSTEM\CurrentControlSet\Services\Tcpip\Parameters\Interfaces')
    i = 0
    while True:
        try:
            guid = winreg.EnumKey(interfaces, i)
            i += 1
            if not _nt_is_enabled(hlm, guid):
                continue
            servers = _nt_read_key(interfaces, guid)
            if servers is not None:
                nameservers.extend(servers)
        except EnvironmentError:
            break
    interfaces.Close()
    hlm.Close()
    return nameservers



from __future__ import annotations
import struct
from typing import Dict, Iterable, Tuple
from .record_data import RecordData
from .record_types import RecordType
from .utils import (
    load_domain_name,
    pack_domain_name
)


class NAPTRRecordData(RecordData):
    '''A record'''

    def __init__(self, *args):
        super().__init__(
            RecordType.SRV,
            data=args
        )

        (
            order,
            preference,
            flags,
            service,
            regexp,
            replacement
        ) = args

        self.order = order
        self.preference = preference
        self.flags = flags
        self.service = service
        self.regexp = regexp
        self.replacement = replacement

    def __repr__(self):
        return '<%s-%s-%s: %s %s %s %s>' % (
            self.type_name, 
            self.order, 
            self.preference, 
            self.flags,
            self.service, 
            self.regexp, 
            self.replacement
        )


    @classmethod
    def load(
        cls, 
        data: bytes, 
        cursor_position: int, 
        size: int
    ) -> Tuple[int, NAPTRRecordData]:
        pos = cursor_position

        order, preference = struct.unpack('!HH', data[pos:pos + 4])
        pos += 4
        
        length = data[pos]
        pos += 1

        flags = data[pos:pos + length].decode()
        pos += length

        length = data[pos]
        pos += 1

        service = data[pos:pos + length].decode()
        pos += length

        length = data[pos]
        pos += 1

        regexp = data[pos:pos + length].decode()
        pos += length

        cursor_position, replacement = load_domain_name(data, pos)
        return cursor_position, NAPTRRecordData(
            order, 
            preference, 
            flags, 
            service, 
            regexp, 
            replacement
        )

    def dump(self, names: Dict[str, int], offset: int) -> Iterable[bytes]:
        raise NotImplementedError





import asyncio
from typing import List, Tuple, Union
from mercury_sync.discovery.dns.core.url import (
    URL,
    InvalidHost,
    InvalidIP

)
from mercury_sync.discovery.dns.core.cache import CacheNode
from mercury_sync.env import (
    Env, 
    RegistrarEnv, 
    load_env
)
from mercury_sync.env.time_parser import TimeParser
from mercury_sync.models.dns_entry import DNSEntry
from mercury_sync.models.dns_message import (
    DNSMessage
)
from mercury_sync.discovery.dns.core.exceptions import (
    DNSError
)
from mercury_sync.discovery.dns.core.record import RecordType
from mercury_sync.discovery.dns.core.record.record_data_types import (
    CNAMERecordData,
    NSRecordData
)
from mercury_sync.discovery.dns.request.dns_client import DNSClient
from .memoizer import Memoizer


class BaseResolver:
    zone_domains = []
    nameserver_types = [
        RecordType.A
    ]
    memoizer=Memoizer()

    def __init__(    
        self,
        host: str,
        port: int,
        instance_id: str,
        env: Env,
        cache: CacheNode = None
    ):
        self.host = host
        self.port = port
        self._queries = {}
        self.cache = cache or CacheNode()
        self.client = DNSClient(
            host,
            port,
            instance_id,
            env
        )

        registrar_env: RegistrarEnv = load_env(RegistrarEnv)
        
        self._request_timeout = TimeParser(
            registrar_env.MERCURY_SYNC_RESOLVER_REQUEST_TIMEOUT
        ).time

    def cache_message(self, query: DNSEntry):
        for _, record in query.to_record_data():
            if query.time_to_live > 0 and record.rtype != RecordType.SOA:
                self.cache.add(record=record)

    def set_zone_domains(self, domains: List[str]):
        self.zone_domains = [
            domain.lstrip('.') for domain in domains
        ]

    async def _query(self, _fqdn: str, _record_type: RecordType) -> DNSMessage:
        raise NotImplementedError

    async def query(
        self,
        fqdn: str,
        record_type: RecordType=RecordType.ANY,
        skip_cache: bool=False
    ):
        
        if fqdn.endswith('.'):
            fqdn = fqdn[:-1]

        if record_type == RecordType.ANY:
            try:
                addr = URL(
                    fqdn,
                    port=self.port
                )

                ptr_name = addr.to_ptr()

            except (InvalidHost, InvalidIP):
                pass

            else:
                fqdn = ptr_name
                record_type = RecordType.PTR

        try:

            return await asyncio.wait_for(
                self._query(
                    fqdn, 
                    record_type,
                    skip_cache
                ),
                timeout=self._request_timeout
            )
        
        except asyncio.TimeoutError:
            return DNSMessage()

    async def request(
        self, 
        fqdn: str, 
        message: DNSMessage, 
        url: URL
    ) -> DNSMessage:

        result = await self.client.send(fqdn, message, url)

        if len(result.query_domains) < 1:
            return False, fqdn, []

        if result.query_domains[0].name != fqdn:
            raise DNSError(-1, 'Question section mismatch')
        
        assert result.query_result_code != 2, 'Remote server fail'

        self.cache_message(result)

        return result

    def _add_cache_cname(self, msg: DNSMessage, fqdn: str) -> Union[str, None]:
        for cname in self.cache.query(fqdn, RecordType.CNAME):
            msg.query_answers.append(cname.copy(name=fqdn))
            if isinstance(cname.data, CNAMERecordData):
                return cname.data.data

    def _add_cache_cname(
        self, 
        msg: DNSMessage, 
        fqdn: str
    ) -> Union[str, None]:
        
        for cname in self.cache.query(fqdn, RecordType.CNAME):
            msg.query_answers.append(cname.copy(name=fqdn))
            if isinstance(cname.data, CNAMERecordData):
                return cname.data.data
            
    def _add_cache_qtype(
        self, 
        msg: DNSMessage, 
        fqdn: str, 
        record_type: RecordType
    ) -> bool:
        if record_type == RecordType.CNAME:
            return False
        
        has_result = False
        for rec in self.cache.query(fqdn, record_type):

            if isinstance(rec.data, NSRecordData):
                a_res = list(self.cache.query(
                    rec.data.data, (
                        RecordType.A,
                        RecordType.AAAA
                    )
                ))
                
                if a_res:
                    msg.query_additional_records.extend(a_res)
                    msg.query_namservers.append(rec)
                    has_result = True
            else:
                msg.query_answers.append(rec.copy(name=fqdn))
                has_result = True

        return has_result

    def _add_cache_record_type(
        self, 
        msg: DNSMessage, 
        fqdn: str, 
        record_type: RecordType
    ) -> bool:
        
        if record_type == RecordType.CNAME:
            return False
        
        has_result = False
        for rec in self.cache.query(fqdn, record_type):

            if isinstance(rec.data, NSRecordData):
                records = list(self.cache.query(
                    rec.data.data, (
                        RecordType.A,
                        RecordType.AAAA
                    )
                ))
                
                if records:
                    msg.query_additional_records.extend(records)
                    msg.query_namservers.append(rec)
                    has_result = True
            else:
                msg.query_answers.append(rec.copy(name=fqdn))
                has_result = True

        return has_result
 
    def query_cache(
        self, 
        msg: DNSMessage,
        fqdn: str, 
        record_type: RecordType
    ):

        cnames = set()

        while True:
            cname = self._add_cache_cname(msg, fqdn)
            if not cname: 
                break

            if cname in cnames:
                # CNAME cycle detected
                break

            cnames.add(cname)
            # RFC1034: If a CNAME RR is present at a node, no other data should be present
            fqdn = cname

        has_result = bool(cname) and record_type in (RecordType.CNAME, RecordType.ANY)

        if record_type != RecordType.CNAME:
            has_result = self._add_cache_qtype(msg, fqdn, record_type) or has_result

        if any(('.' + fqdn).endswith(root) for root in self.zone_domains):
            if not has_result:
                msg.query_result_code = 3
                has_result = True

            msg.query_authoritative_answer = 1
        # fqdn may change due to CNAME
        return has_result, fqdn


from __future__ import annotations
import re
from mercury_sync.discovery.dns.core.exceptions import InvalidServiceURLError
from mercury_sync.discovery.dns.core.record.record_data_types import (
    ARecordData,
    AAAARecordData,
    CNAMERecordData,
    PTRRecordData,
    SRVRecordData,
    TXTRecordData,
    RecordType
)
from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictFloat,
    IPvAnyAddress
)

from typing import (
    Literal, 
    Tuple, 
    Optional, 
    Dict,
    Union,
    List
)



DomainProtocol = Literal["tcp", "udp"]
RecordTypeName = Literal["A", "AAAA", "CNAME", "PTR", "SRV", "TXT"]


service_pattern = re.compile(r'([a-zA-Z0-9\-]{1,256})?(\.?\_)([a-zA-Z0-9\-]{1,256})(\._)([udp|tcp]*)(\.)([a-zA-Z0-9\-]{1,256})(\.)([a-zA-Z0-9]{2,5})')
ptr_service_pattern = re.compile(r'([a-zA-Z0-9\-]{1,256})(\._)([udp|tcp]*)(\.)([a-zA-Z0-9\-]{1,256})(\.)([a-zA-Z0-9]{2,5})')


class DNSEntry(BaseModel):
    instance_name: Optional[StrictStr]
    service_name: StrictStr
    domain_protocol: DomainProtocol
    domain_name: StrictStr
    domain_priority: StrictInt=10
    domain_weight: StrictInt=0
    domain_port: Optional[StrictInt]
    domain_values: Dict[StrictStr, StrictStr]={}
    domain_targets: Optional[
        Tuple[
            Union[IPvAnyAddress, StrictStr]
        ]
    ]
    record_type: Optional[RecordType]
    record_types: List[RecordTypeName]=["PTR", "SRV", "TXT"]
    time_to_live: Union[StrictInt, StrictFloat]=-1


    @classmethod
    def to_segments(cls, url: str):

        if service_pattern.match(url) is None:
            raise InvalidServiceURLError(url)

        segments = [
            segment for segment in service_pattern.split(url) if segment.isalnum()
        ]

        instance_name, service_name, domain_protocol = segments[:3]
        domain_name = '.'.join(segments[3:])

        return (
            instance_name,
            service_name,
            domain_protocol,
            domain_name
        )
    
    @classmethod
    def to_ptr_segments(cls, url: str):

        if ptr_service_pattern.match(url) is None:
            raise InvalidServiceURLError(url)

        segments =  [
            segment for segment in ptr_service_pattern.split(url) if segment.isalnum()
        ]

        service_name, domain_protocol = segments[:2]
        domain_name = '.'.join(segments[2:])

        return (
            service_name,
            domain_protocol,
            domain_name
        )

    def to_domain(
        self,
        record_type: RecordTypeName
    ):

        if record_type == "PTR":
            domain = f'{self.service_name}._{self.domain_protocol}.in-addr.arpa'

        else:
             domain = f'{self.instance_name}._{self.service_name}._{self.domain_protocol}.{self.domain_name}'

        return domain


    def to_data(
        self,
        record_type: RecordTypeName
    ):

        domain_target: Union[str, None] = None

        if self.domain_targets:
            domain_target = str(self.domain_targets[0])

        if record_type == "A":
            return ARecordData(domain_target)
        
        elif record_type == "AAAA":
            return AAAARecordData(domain_target)
        
        elif record_type == "CNAME":
            return CNAMERecordData(domain_target)
        
        elif record_type == "SRV":
            return SRVRecordData(
                self.domain_priority,
                self.domain_weight,
                self.domain_port,
                domain_target
            )
        
        elif record_type == "PTR" and self.instance_name:
            domain_target = f'{self.instance_name}._{self.service_name}._{self.domain_protocol}.{self.domain_name}'
            return PTRRecordData(domain_target)
        
        elif record_type == "PTR":
            domain_target = f'{self.instance_name}._{self.service_name}._{self.domain_protocol}.{self.domain_name}'
            return PTRRecordData(domain_target)
        
        else:
            domain_target_value = f'service={domain_target}'
            txt_values = [
                f'{key}={value}' for key, value in self.domain_values.items()
            ]

            txt_values.append(domain_target_value)

            txt_record_data = '\n'.join(txt_values)

            return TXTRecordData(txt_record_data)
        
    def to_record_data(self) -> List[
        Tuple[
            str, 
            Union[
                ARecordData,
                AAAARecordData,
                CNAMERecordData,
                PTRRecordData,
                SRVRecordData,
                TXTRecordData
            ]
        ]
    ]:
        return [
            (
                self.to_domain(record_type),
                self.to_data(record_type)
            ) for record_type in self.record_types
        ]
        
    @classmethod
    def from_record_data(
        self,
        record_name: str,
        record_data: Union[
            ARecordData,
            AAAARecordData,
            CNAMERecordData,
            SRVRecordData,
            TXTRecordData
        ]
    ):
        
        if record_data.rtype == RecordType.PTR:
            (
                instance_name, 
                service_name, 
                domain_protocol,
                domain_name
            ) = DNSEntry.to_segments(record_data.data)

        else:
            
            (
                instance_name, 
                service_name, 
                domain_protocol,
                domain_name
            ) = DNSEntry.to_segments(record_name)
        
        if isinstance(
            record_data, 
            (
                ARecordData, 
                AAAARecordData, 
                CNAMERecordData
            )
        ):
            return DNSEntry(
                instance_name=instance_name,
                service_name=service_name,
                domain_protocol=domain_protocol,
                domain_name=record_name,
                domain_targets=(
                    record_data.data,
                ),
                record_type=record_data.rtype
            )
        
        elif isinstance(record_data, PTRRecordData):

            return DNSEntry(
                instance_name=instance_name,
                service_name=service_name,
                domain_protocol=domain_protocol,
                domain_name=domain_name,
                domain_targets=(
                    record_data.data,
                ),
                record_type=record_data.rtype
            )
        
        elif isinstance(record_data, SRVRecordData):

            return DNSEntry(
                instance_name=instance_name,
                service_name=service_name,
                domain_protocol=domain_protocol,
                domain_name=domain_name,
                domain_port=record_data.port,
                domain_priority=record_data.priority,
                domain_weight=record_data.weight,
                domain_targets=(
                    record_data.hostname,
                ),
                record_type=record_data.rtype
            )
        
        else:

            txt_data = record_data.data.split("\n")

            record_values: Dict[str, str] = {}

            for txt_item in txt_data:
                key, value = txt_item.split("=")
                record_values[key] = value

            domain_target = record_values.get("service")

            return DNSEntry(
                instance_name=instance_name,
                service_name=service_name,
                domain_protocol=domain_protocol,
                domain_name=record_name,
                domain_targets=(
                    domain_target,
                ),
                domain_values=record_values,
                record_type=record_data.rtype
            )





from pathlib import Path


def get_nameservers(filename='/etc/resolv.conf'):

    nameservers = []

    for line in Path(filename).read_text().splitlines():

        if line.startswith('#'):
            continue

        parts = line.split()
        if len(parts) < 2:
            continue

        if parts[0] == 'nameserver':
            nameservers.append(parts[1])

    return nameservers



from typing import List, Iterable, Generic, TypeVarTuple, Union
from .service import Service


P = TypeVarTuple('P')

class PluginGroup(Generic[*P]):

    def __init__(
        self,
        service_pool: List[Union[*P]]
    ) -> None:
        self._services = service_pool
        self._services_count = len(service_pool)
        self._current_idx = 0

    @property
    def one(self) -> Union[*P]:
        service: Service = self._services[self._current_idx]
        self._current_idx = (self._current_idx + 1)%self._services_count

        return service
    
    def each(self) -> Iterable[Union[*P]]:
        for service in self._services:
            yield service
    
    def at(self, idx: int) -> Union[*P]:
        return self._services[idx]



import functools
from pydantic import BaseModel
from typing import (
    Optional, 
    List, 
    Literal,
    Dict,
    Callable
)


def endpoint(
    path: Optional[str]="/",
    methods: List[
        Literal[
            "GET",
            "HEAD",
            "OPTIONS",
            "POST",
            "PUT",
            "PATCH",
            "DELETE"
        ]
    ]=["GET"],
    responses: Optional[
        Dict[
            int,
            BaseModel
        ]
    ]=None,
    serializers: Optional[
        Dict[
            int,
            Callable[
                ...,
                str
            ]
        ]
    ]=None,
    response_headers: Optional[Dict[str, str]]=None
):

    def wraps(func):

        func.server_only = True
        func.path = path
        func.methods = methods
        func.as_http = True
        func.response_headers = response_headers
        func.responses = responses
        func.serializers = serializers

        @functools.wraps(func)
        def decorator(
            *args,
            **kwargs
        ):
            return func(*args, **kwargs)
        
        return decorator
    
    return wraps
